@@ -5,8 +5,9 @@ use Mojo::Base 'DBI';
 # Register subclass structure
 __PACKAGE__->init_rootclass;
 
-our $VERSION = 2.113;
+our $VERSION = 2.161;
 
+use Carp 'croak';
 use File::Spec::Functions 'catfile';
 use Mojar::ClassShare 'have';
 
@@ -63,25 +64,27 @@ have Defaults => sub { bless {} => ref $_[0] || $_[0] };
 
 # Attributes
 
-my @DbdFields = qw( RaiseError PrintError PrintWarn AutoCommit TraceLevel
-    mysql_enable_utf8 mysql_auto_reconnect );
+has quiesce_timeout => 500;
+
+my @DbdFields = qw(RaiseError PrintError PrintWarn AutoCommit TraceLevel
+    mysql_auto_reconnect mysql_enable_utf8);
 
 has RaiseError => 1;
 has PrintError => 0;
 has PrintWarn => 0;
 has AutoCommit => 1;
 has TraceLevel => 0;
-has mysql_enable_utf8 => 1;
 has mysql_auto_reconnect => 0;
+has mysql_enable_utf8 => 1;
 
-my @ConFields = qw( label cnfdir cnf cnfgroup );
+my @ConFields = qw(label cnfdir cnf cnfgroup);
 
 has 'label';
 has cnfdir => '.';
 has 'cnf';
 has 'cnfgroup';
 
-my @DbiFields = qw( driver host port schema user password );
+my @DbiFields = qw(driver host port schema user password);
 
 has driver => 'mysql';
 has 'host';  # eg 'localhost'
@@ -90,10 +93,6 @@ has 'schema';  # eg 'test';
 has 'user';
 has 'password';
 
-# Private function
-
-sub croak { require Carp; goto &Carp::croak; }
-
 # Public methods
 
 sub new {
@@ -101,7 +100,8 @@ sub new {
   # $proto may contain defaults to be cloned
   # %param may contain defaults for overriding
   my %defaults = ref $proto ? ( %{ ref($proto)->Defaults }, %$proto )
-                            : %{ $proto->Defaults };
+                            : %{$proto->Defaults};
+  delete $defaults{$_} for grep { ref $proto and /^dbh\./ } keys %defaults;
   return Mojo::Base::new($proto, %defaults, %param);
 }
 
@@ -110,17 +110,20 @@ sub connect {
   my $class = ref $proto || $proto;
   @args = $proto->dsn(@args) unless @args and $args[0] =~ /^DBI:/i;
   my $dbh;
-  eval {
-    $dbh = $class->SUPER::connect(@args)
-  }
+  eval { $dbh = $class->SUPER::connect(@args) }
   or do {
     my $e = $@;
-    croak sprintf "Connection error\n%s\n%s",
-        $proto->dsn_to_dump(@args), $e;
+    croak sprintf "Connection error\n%s\n%s", $proto->dsn_to_dump(@args), $e;
   };
   return $dbh;
 }
 
+sub connection {
+  my ($self, $tag) = @_; $tag //= 'connection';
+  return $self->{"dbh.$tag"} if ($self->{"dbh.$tag"} //= $self->connect)->ping;
+  return $self->{"dbh.$tag"} = $self->connect;
+}
+
 sub dsn {
   my ($proto, %param) = @_;
   my $param = $proto->new(%param);
@@ -142,6 +145,7 @@ sub dsn {
   my %custom;
   defined($param->$_) and $custom{$_} = $param->$_ for qw(label cnf cnfgroup);
   my $dbd_param = %custom ? { private_config => {%custom} } : {};
+  $dbd_param->{$_} = $param->{$_} for grep /^mysql_/, keys %$param;
   @$dbd_param{@DbdFields} = map $param->$_, @DbdFields;
 
   return (
@@ -173,13 +177,10 @@ sub dsn_to_dump {
 package Mojar::Mysql::Connector::db;
 @Mojar::Mysql::Connector::db::ISA = 'DBI::db';
 
+use Carp 'croak';
 use Mojar::Util 'lc_keys';
 use Scalar::Util 'looks_like_number';
 
-# Private functions
-
-sub croak { require Carp; goto &Carp::croak; }
-
 our $_as_hash = { Slice => {} };
 sub as_hash { $_as_hash }
 
@@ -217,7 +218,7 @@ sub session_var { shift->_var('SESSION', @_) }
 
 sub global_var {
   my $self = shift;
-  return $self->_var('GLOBAL', @_)
+  return $self->_var(GLOBAL => @_)
     if @_ >= 2 or @_ == 1 and $_[0] ne 'have_innodb';
 
   my $variables = $self->_var('GLOBAL');
@@ -231,20 +232,20 @@ sub global_var {
   return $variables;
 }
 
-sub disable_quotes { shift->session_var( sql_quote_show_create => 0 ) }
+sub disable_quotes { shift->session_var(sql_quote_show_create => 0) }
 
 sub enable_quotes {
   my ($self, $value) = @_;
   $value //= 1;
-  $self->session_var( sql_quote_show_create => $value )
+  $self->session_var(sql_quote_show_create => $value)
 }
 
-sub disable_fk_checks { shift->session_var( foreign_key_checks => 0 ) }
+sub disable_fk_checks { shift->session_var(foreign_key_checks => 0) }
 
 sub enable_fk_checks {
   my ($self, $value) = @_;
   $value //= 1;
-  $self->session_var( foreign_key_checks => $value )
+  $self->session_var(foreign_key_checks => $value)
 }
 
 sub schemata {
@@ -255,7 +256,7 @@ sub schemata {
     my $sql = q{SHOW DATABASES};
     $sql .= sprintf q{ LIKE '%s'}, $args[0] if defined $args[0];
     $schemata = $self->selectcol_arrayref($sql, $args[1]) or die;
-    @$schemata = grep !/^lost\+found/, @$schemata;
+    @$schemata = grep !/^(?:\#|lost\+found)/, @$schemata;
     1;
   }
   or do {
@@ -312,7 +313,14 @@ sub selectall_arrayref_hashrefs {
   return $self->selectall_arrayref($sql, $opts, @args);
 }
 
-sub processes {
+sub selectall_lookup {
+  my $self = shift;
+  my $rs = $self->selectall_arrayref(@_);  # ($sql, $opts, @args)
+  return undef if @$rs >= 1 and @{$$rs[0]} != 2;  # wrong qty cols
+  return {map @$_, @$rs};  # flatten
+}
+
+sub threads {
   my $p = shift->selectall_arrayref_hashrefs(q{SHOW FULL PROCESSLIST});
   @$p = map lc_keys($_), @$p;
   return $p;
@@ -345,6 +353,24 @@ sub statistics {
   return lc_keys { map @$_, @$s };
 }
 
+sub engine {
+  my ($self, $schema, $table) = @_;
+  my $engine;
+  if ($self->mysqld_version =~ /^(\d+)\./ and $1 >= 5) {
+    ($engine) = $self->selectrow_array(
+q{SELECT ENGINE
+FROM information_schema.TABLES
+WHERE
+  TABLE_SCHEMA = ?
+  AND TABLE_NAME = ?},
+      undef,
+      $schema,
+      $table
+    );
+  }
+  return $engine;
+}
+
 sub indices {
   my ($self, $schema, $table) = @_;
   croak 'Missing required schema name' unless defined $schema and length $schema;
@@ -373,30 +399,37 @@ q{SHOW TABLE STATUS FROM %s}, $schema;
 sub engine_status {
   my ($self, $engine) = @_;
   $engine //= 'InnoDB';
+  Carp::croak "Bad engine ($engine)" unless $engine =~ /^\w+$/;
 
-  my ($raw) = $self->selectrow_array(
-q{SHOW INNODB STATUS}
-  );
+  my ($major, $minor) = ($self->mysqld_version =~ /^(\d+)\.(\d+)\./);
+  my $i = $major > 5 ? 2 : $major == 5 && $minor >= 1 ? 2 : 0;
+  my $raw = ($self->selectrow_array(
+      qq{SHOW /*!50500 ENGINE */ $engine STATUS}))[$i]
+      // die "Failed to get engine ($engine) status\n";
+  $raw =~ s/\t/ /g;
 
-  my ($title, $buffer) = ('', '');
-  my $status = {};
-  for (split /^/, $raw) {
-    if (/^\-+$/ and length $buffer) {
+  my ($title, $buffer, $status) = ('', '', {});
+  while ($raw =~ /^(.*)$/mg) {
+    my $line = $1;
+    if ($line =~ /^-+$/ and length $buffer) {
       # Finish previous record
       $status->{$title} = $buffer;
       $title = $buffer = '';
     }
-    elsif (/^-+$/) {
+    elsif ($line =~ /^-+$/) {
       # Start new record
     }
     elsif (not length $title) {
-      chomp;
-      $title = lc $_;
-      $title =~ s/\s/_/g;
+      chomp $line;
+      $title = lc $line;
+      $title =~ s/\s+/_/g;
       $title =~ s/\W//g;
     }
+    elsif ($line =~ /^=+$/) {
+      next;
+    }
     else {
-      $buffer .= $_;
+      $buffer .= $line . $/;
     }
     # Ignore final record
   }
@@ -436,6 +469,46 @@ q{SELECT DATE_FORMAT(DATE_ADD(CURDATE(), INTERVAL %s DAY), '%s')},
   return $date;
 }
 
+sub quiesce {
+  my ($self) = @_;
+
+  # Record existing state
+  my $pct = $self->global_var('innodb_max_dirty_pages_pct')
+    // die "Failed to get dirty_pages info\n";
+  my $repl = $self->selectrow_hashref('SHOW SLAVE STATUS') // {};
+  Carp::croak "Cannot quiesce while replicating"
+    if ($repl->{Slave_SQL_Running} // '') =~ /^Yes/i;
+
+  # Prepare callback to un-quiesce db
+  my $cb = sub {
+    eval { $self->do('UNLOCK TABLES') };
+    $self->global_var(innodb_max_dirty_pages_pct => shift // $pct);
+  };
+
+  # Quiesce
+  eval {
+    $self->global_var(innodb_max_dirty_pages_pct => 0);
+    $self->do('FLUSH TABLES WITH READ LOCK');
+    my ($quiesced, $count) = (0, 0);
+    while (++$count < $self->quiesce_timeout) {
+      my $state = $self->engine_status('InnoDB')->{buffer_pool_and_memory};
+      if ($state =~ /^Modified db pages\s+(\d+)$/m) {
+        ++$quiesced, last if $1 == 0;
+      }
+      else {
+        die "Failed to check dirty pages\n";
+      }
+      sleep 1;
+    }
+    $quiesced;
+  } or do {
+    my $e = $@ // '';
+    eval { $cb->(); };
+    die "Failed to quiesce database: $e\n";
+  };
+  return $cb;
+}
+
 # Private method
 
 sub _var {
@@ -700,11 +773,38 @@ All connector parameters are implemented as attributes with exactly the same
 spelling.  So for example you can
 
   $connector->RaiseError(undef);  # disable RaiseError
-  $connector->mysql_enable_utf8(1);  # enable mysql_enable_utf8
+  $connector->mysql_enable_utf8(0);  # disable mysql_enable_utf8
+
+The attributes, with their coded defaults, are
+
+  RaiseError => 1
+  PrintError => 0
+  PrintWarn => 0
+  AutoCommit => 1
+  TraceLevel => 0
+  mysql_auto_reconnect => 0
+  mysql_enable_utf8 => 1
+
+  label
+  cnfdir => '.'
+  cnf
+  cnfgroup
+
+  driver => 'mysql'
+  host
+  port
+  schema
+  user
+  password
+
+In addition, any L<DBD::mysql> attributes (beginning "mysql_") are passed
+through to the driver.
+
+  $dbh = $connector->connect(mysql_skip_secure_auth => 1);
 
 =head1 DATABASE HANDLE METHODS
 
-=head2 C<selectall_arrayref_hashref>
+=head2 C<selectall_arrayref_hashrefs>
 
   $_->{Command} ne 'Sleep' and say $_->{User}
     for $dbh->selectall_arrayref_hashrefs(q{SHOW FULL PROCESSLIST});
@@ -825,6 +925,38 @@ Returns a arrayref of view names, similar to
 
 but excluding real tables.
 
+=head1 CHARACTER ENCODINGS
+
+To read/store characters encoded as non-ASCII, non-UTF8, you must disable
+handling of UTF-8.
+
+  $connector = Mojar::Mysql::Connector->new(mysql_enable_utf8 => 0);
+
+This is essential, for example, when fetching high-latin (eg non-ASCII 8859-1)
+characters.
+
+=head1 DEBUGGING
+
+You can enable DBI trace logging at use-time:
+
+  use Mojar::Mysql::Connector (TraceLevel => '3|CON');
+
+and since you have access to all of L<DBI>, you can set tracing using the method
+
+  $dbh->trace('3|CON');
+  ...
+  $dbh->trace(0);
+
+or by using the attribute
+
+  {
+    local $dbh->{TraceLevel} = '3|CON';
+    ...
+  }
+
+To set tracing for all handles, use the class method instead.  See
+L<DBI/TRACING>.
+
 =head1 SUPPORT
 
 =head2 Homepage
@@ -852,13 +984,19 @@ alternatives out there but I'm still surprised how little support there is for
 keeping passwords out of your codebase and helping you manage multiple
 connections.
 
-=head1 SEE ALSO
-
-L<Coro::Mysql>, L<AnyEvent::DBI>, L<DBIx::Custom>, L<DBIx::Connector>, L<DBI>.
-
 =head1 COPYRIGHT AND LICENCE
 
-Copyright (C) 2002--2014, Nic Sandfield.
+Copyright (C) 2002--2017, Nic Sandfield.
 
 This program is free software, you can redistribute it and/or modify it under
 the terms of the Artistic License version 2.0.
+
+=head1 SEE ALSO
+
+L<Mojo::mysql>, L<Coro::Mysql>, L<AnyEvent::DBI>, L<DBIx::Custom>,
+L<DBIx::Connector>, L<DBI>.
+
+=cut
+
+# engine_status and quiesce have been tested on 4.0.27, 4.1.11, 5.0.87, 5.1.73,
+# 5.6.23, 5.7.16
